#version 330 core

// Input vertex data, different for all executions of this shader.
layout(location = 0) in vec3 vPosition; // vertex position (in model space) gl_Vertex is removed from version140
layout(location = 1) in vec2 vTexCoord; // texture coordinate (in model space) gl_Normal is removed from version140
layout(location = 2) in vec3 vNormal; // vertex normal (in model space) gl_MultiTexCoord0 is removed from version140

// Output data will be interpolated for each fragment.
out vec3 PositionWorldPass;
out vec2 TexCoordPass;
out vec3 NormalWorldPass;

// Values that stay constant for the whole mesh.
uniform mat4 gModel, gView, gProjection;

void main()
{
	// Output position of the vertex, in clip space : gMVP * position
	mat4 MVP = gProjection * gView * gModel;
	gl_Position =  MVP * vec4(vPosition,1);
	
	// Position of the vertex, in worldspace : gWorld * position
	PositionWorldPass = (gModel * vec4(vPosition,1)).xyz;
	
	// Vertex normal vector in world space
	NormalWorldPass = (gModel * vec4(vNormal, 1)).xyz;

	// UV of the vertex. No special space for this one.
	TexCoordPass = vTexCoord;
}

