#include "moglclass.h"
#include "shaders.h"
#include "geometryPositionNormalTexture.h"
#include "light.h"
#include "material.h"
#include "texture2D.h"
#include "camera.h"

using namespace moglclass;

#define NUMSHADERS 4

Shader shShaders[NUMSHADERS];
Program spMain;

float g_fovy = 45.0f;
float g_aspect = 1.0f;
float g_zNear = 1.0f;
float g_zFar = 100.f;

Camera camera1(FLY);
//float cameraHeading = 0.0f;
//float cameraZ = 10.0f;
glm::vec3 g_eye(0, 0, 6);
glm::vec3 g_at(0, 0, 0);
glm::vec3 g_up(0, 1, 0);

glm::mat4 Identity(1.0f);
glm::mat4 World(1.0f);
glm::mat4 View(1.0f);
glm::mat4 Projection(1.0f);
glm::mat4 MVP(1.0f);
glm::mat4 xyRotation(1.0f);

Light light0(glm::vec4(0.0f, 0.0f, 5.0f, 0.0f), glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(0.5f, 0.5f, 0.5f, 0.5f), 1.0f, 0.0f, 0.0f, 180.0f, 0.0f, glm::vec3(0.0f, 0.0f, 0.0f)); 
Light light1(glm::vec4(-1.0f, 1.0f, 0.0f, 0.0f), glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 1.0f, 0.0f, 0.0f, 180.0f, 0.0f, glm::vec3(0.0f, 0.0f, 0.0f));
Light light2(glm::vec4(-5.0f, -5.0f, 4.0f, 0.0f), glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 0.0f, 1.0f, 0.0f, 60.0f, 50.0f, glm::vec3(0.0f, 0.0f, -1.0f));

Material whiteMaterial(glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 50.0f);
float g_alpha = 1.0f;

Texture2D * mainTex = NULL;
Texture2D * princessTex = NULL;
Texture2D * openglTex = NULL;

Quad * wall = NULL;
Quad * rect = NULL;
Cylinder * cyl = NULL;
Sphere * sph = NULL;

bool g_wiremode = false;
bool g_animmode = false;
int g_drawmode = 0;
int g_filter = 0;
glm::vec3 g_SpritePosition(-3, 0, 2);
bool g_fademode = false;
int startTime = 0;

void init( void )
{
    // load shaders and use the resulting shader program
	spMain.createProgram();
	shShaders[0].loadShader("MTexturePerPixelShading.vs", GL_VERTEX_SHADER);
	shShaders[1].loadShader("SimpleTextureShading.fs", GL_FRAGMENT_SHADER);
	spMain.attachShaderToProgram(&shShaders[0]);
	spMain.attachShaderToProgram(&shShaders[1]);
	spMain.linkProgram();
	spMain.useProgram();
	
	Projection = glm::perspective(g_fovy, g_aspect, g_zNear, g_zFar);
	View = camera1.lookAt(g_eye, g_at, g_up);

	// texture
	mainTex = new Texture2D("main.jpg");
	princessTex = new Texture2D("danc_princess.png");
	openglTex = new Texture2D("opengl.jpg");
	
	wall = new Quad(glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f), glm::vec3(0.0f, 1.0f, 0.0f), 8.0f, 6.0f); // texture quad
	rect = new Quad(glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f), glm::vec3(0.0f, 1.0f, 0.0f), 2.0f, 2.0f); // texture quad
	cyl = new Cylinder(glm::vec3(-1.0, 0, -0.5), 1.0f, 4.0f, 16);
	sph = new Sphere(glm::vec3(1.5, 0, 0.5), 1.5f, 16, 16);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glClearColor( 0.7, 0.8, 1.0, 1.0 ); // white background
}

void cleanup()
{
	if (wall) delete wall;
	if (rect) delete rect;
	if (cyl) delete cyl;
	if (sph) delete sph;

	spMain.deleteProgram();
	printf("clean up\n");
}

void drawTextureBackground(glm::mat4& worldMatrix, glm::vec3& position)
{
	//World = glm::translate(glm::mat4(1.0f), position);
	World = glm::translate(worldMatrix, position);
	spMain.setUniform("gModel", World);
	spMain.setUniform("gTextureSampler", 0);
	mainTex->bind();
	wall->draw();
	mainTex->disable();
}

void drawTextureSprite(glm::mat4& worldMatrix, glm::vec3& position)
{
	//World = glm::translate(glm::mat4(1.0f), position);
	World = glm::translate(worldMatrix, position);
	spMain.setUniform("gModel", World);
	spMain.setUniform("gTextureSampler", 0);
	princessTex->bind();
	rect->draw();
	princessTex->disable();
}

void drawTextureCylinderSphere(glm::mat4& worldMatrix, glm::vec3& position)
{
	//World = glm::translate(glm::mat4(1.0f), position);
	World = glm::translate(worldMatrix, position);
	spMain.setUniform("gModel", World);
	spMain.setUniform("gTextureSampler", 0);
	mainTex->bind();
	cyl->draw();
	mainTex->disable();
	mainTex->bind();
	sph->draw();
	mainTex->disable();
}

void drawTextureBlendFilter(glm::mat4& worldMatrix, glm::vec3& position)
{
	// material (50% transparency)
	whiteMaterial.setDiffuse(glm::vec4(1, 1, 1, 0.5));
	whiteMaterial.setUniform(&spMain, "gMaterial");

	//World = glm::translate(glm::mat4(1.0f), position);
	World = glm::translate(worldMatrix, position);
	spMain.setUniform("gModel", World);
	spMain.setUniform("gTextureSampler", 0);
	// blended tex square
	glEnable(GL_CULL_FACE);
	glEnable(GL_BLEND);

	if (g_filter == 0)                     // Default (no blending) = Cs*1
		glBlendFunc(GL_ONE, GL_ZERO);
	else if (g_filter == 1)                // Draw background only = Cd*1
		glBlendFunc(GL_ZERO, GL_ONE);
	else if (g_filter == 2)                // Cs*1 + Cd*1 
		glBlendFunc(GL_ONE, GL_ONE);
	else if (g_filter == 3)                // Alpha blending (back-to-front) = Cs*As + Cd*(1-As)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	else if (g_filter == 4)                // Additive blending (brighten the scene) = Cs*As + Cd*1
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	else if (g_filter == 5)                // Multiplicative blending = Cd*Cs
		glBlendFunc(GL_ZERO, GL_SRC_COLOR);
	else if (g_filter == 6)                // darken the scene = Cd*As
		glBlendFunc(GL_ZERO, GL_SRC_ALPHA);
	else if (g_filter == 7)                 // Invert all the colors = Cs*(1-Cd)
		glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ZERO);

	openglTex->bind();
	rect->draw();
	openglTex->disable();
	glDisable(GL_BLEND);
}

void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );     // clear the window

	// set drawmode
	if (g_wiremode) 
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);  // draw wireframe 
	else
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);	// draw filled 

	// MVP matrix
	View = camera1.View();
	//World = glm::rotate(glm::mat4(1.0f), -cameraHeading, glm::vec3(0, 1, 0)) * glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, -cameraZ));

	MVP = Projection * View * World;
	spMain.useProgram();
	spMain.setUniform("gMVP", MVP);
	spMain.setUniform("gProjection", Projection);
	spMain.setUniform("gView", View);
	spMain.setUniform("gModel", World);

	glm::vec3 cameraPos = camera1.getPosition();
	//printf("cameraPos: %f %f %f\n", cameraPos[0], cameraPos[1], cameraPos[2]);
	spMain.setUniform("gEyePosition", cameraPos);

	// light
	light0.setUniform(&spMain, "gLight0");
	//light1.setUniform(&spMain, "gLight1");
	//light2.setUniform(&spMain, "gLight2");

	// material
	whiteMaterial.setDiffuse(glm::vec4(1, 1, 1, 1));
	whiteMaterial.setUniform(&spMain, "gMaterial");
	// draw background & blend filter
	drawTextureCylinderSphere(xyRotation, glm::vec3(0, 0, 0));
	drawTextureBlendFilter(xyRotation, glm::vec3(0, 0, 2));
	//drawTextureBackground(xyRotation, glm::vec3(0, 0, 1));
	//drawTextureSprite(xyRotation, g_SpritePosition);

	glutSwapBuffers();
}

void update()
{
	static int currentTime, deltaTime, prevTime = 0; 
	static int elapsedTime = 0;
	
	currentTime = glutGet(GLUT_ELAPSED_TIME);
	deltaTime = currentTime - prevTime;
	prevTime = currentTime;

	if (g_fademode)
	{
		g_alpha = 1.0f - fabs(sinf((float)elapsedTime * 0.0002f));
		printf("g_alpha=%f\n", g_alpha);
	}

	if (g_animmode)
	{
		// translate & rotate around x/y axis
		static float angle = 0.0f;
		angle += 0.001f;

		// geometry: translate & rotate in x & rotate in y
		glm::mat4 T = glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, 1));
		glm::mat4 R = glm::rotate(T, angle, glm::vec3(1, 0, 0));
		xyRotation = glm::rotate(R, angle, glm::vec3(0, 1, 0));
	}

	elapsedTime = currentTime - startTime;
	glutPostRedisplay();
}

void keyboard( unsigned char key, int x, int y )
{
	switch (key)  
	{    
	case 0x1B:    
	case 'q':    
	case 'Q':      
		exit(0);     
	case 'b':
		g_filter = (g_filter + 1) % 8;
		printf("g_filter=%d\n", g_filter);
		break;
	case 'm':
		g_animmode = !g_animmode;
		break;  
	case ' ':
		g_fademode = !g_fademode;
		startTime = glutGet(GLUT_ELAPSED_TIME);
		break;  
	case 'r':
		g_SpritePosition = glm::vec3(0, 0, 2);
		break;
	case 'a':
		g_SpritePosition[0] -= 0.5f;
		break;
	case 'd':
		g_SpritePosition[0] += 0.5f;
		break;
	case 'w':
		g_SpritePosition[1] += 0.5f;
		break;
	case 's':
		g_SpritePosition[1] -= 0.5f;
		break;
	case 'z':
		g_SpritePosition[2] += 0.5f;
		break;
	case 'x':
		g_SpritePosition[2] -= 0.5f;
		break;
	}
	glutPostRedisplay();
}
/*
void specialkey(int key, int x, int y )
{
	switch (key)  
	{    
	case GLUT_KEY_LEFT:
		cameraHeading -= 10;
		printf("cameraHeading=%f\n", cameraHeading);
		break;  
	case GLUT_KEY_RIGHT:    
		cameraHeading += 10;
		printf("cameraHeading=%f\n", cameraHeading);
		break;  
	case GLUT_KEY_UP:      
		cameraZ += 1.0f;
		printf("cameraZ=%f\n", cameraZ);
		break;  
	case GLUT_KEY_DOWN:
		cameraZ -= 1.0f;
		printf("cameraZ=%f\n", cameraZ);
		break;  
	case GLUT_KEY_PAGE_UP:      
		break;  
	case GLUT_KEY_PAGE_DOWN:
		break;  
	}
	glutPostRedisplay();
}
*/

void specialkey(int key, int x, int y)
{
 if (key == GLUT_KEY_F1)   // x-movement
        camera1.strafe(0.5);
 else if (key == GLUT_KEY_F2)
        camera1.strafe(-0.5);
 else if (key == GLUT_KEY_F3) // y-movement
        camera1.fly(0.5);
 else if (key == GLUT_KEY_F4)
        camera1.fly(-0.5);
 else if (key == GLUT_KEY_F5) // z-movement
        camera1.walk(0.5);
    else if (key == GLUT_KEY_F6)
        camera1.walk(-0.5);

 else if (key == GLUT_KEY_F7) // yaw (by y-axis)
        camera1.yaw(2.5);
 else if (key == GLUT_KEY_F8)
        camera1.yaw(-2.5);
 else if (key == GLUT_KEY_F9) // pitch (by x-axis)
        camera1.pitch(2.5);
    else if (key == GLUT_KEY_F10)
        camera1.pitch(-2.5);
 else if (key == GLUT_KEY_HOME) // roll (by z-axis)
        camera1.roll(2.5);
    else if (key == GLUT_KEY_END)
        camera1.roll(-2.5);

 else if (key == GLUT_KEY_LEFT) // same as town
        camera1.yaw(2.5);
 else if (key == GLUT_KEY_RIGHT)
        camera1.yaw(-2.5);
 else if (key == GLUT_KEY_UP) 
        camera1.walk(0.5);
    else if (key == GLUT_KEY_DOWN)
        camera1.walk(-0.5);

 glutPostRedisplay();
}


void mouse(int button, int state, int mx, int my)
{
	if (button==GLUT_RIGHT_BUTTON && state==GLUT_DOWN)
		exit(0);
}

void motion(int mx, int my)
{
	int w = glutGet(GLUT_WINDOW_WIDTH);
	int h = glutGet(GLUT_WINDOW_HEIGHT);
	float x = (float) (mx - w*0.5) / w;
	float y = (float) (h*0.5 - my) / h;
	//fprintf(stderr, "x=%f y=%f\n", x, y);

	glutPostRedisplay();
}

void reshape(int w, int h)
{
	g_aspect = (float) (w/h);

    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    glutPostRedisplay();
}

int main( int argc, char **argv )
{
	glutInit(&argc, argv);
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
    glutInitWindowSize( SCREEN_WIDTH, SCREEN_HEIGHT );
    glutCreateWindow(argv[0]);
    
	glewInit();
    init();

    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
    glutSpecialFunc( specialkey );
	glutMouseFunc( mouse ); 
	glutMotionFunc( motion ); 
	glutIdleFunc( update );
	glutReshapeFunc( reshape );

    glutMainLoop();

	cleanup();

    return 0;
}
